import {IdentifierType, Model, ModelClass, ModelShape} from "@sharkitek/core";
import {Table, TableConfig} from "drizzle-orm";
import {PgDatabase} from "drizzle-orm/pg-core";
import {PgQueryResultHKT} from "drizzle-orm/pg-core/session";
import type {ExtractTablesWithRelations, TablesRelationalConfig} from "drizzle-orm/relations";
import {ModelQuery, query} from "./Query";

/**
 * Sharkitek model extension for Drizzle.
 */
export interface DrizzleExtension<
	ModelType extends Model<Shape, IdentifierType<Shape, Identifier>>,
	TableType extends Table<TC>, TC extends TableConfig,
	TQueryResult extends PgQueryResultHKT, TFullSchema extends Record<string, unknown>,
	Scopes extends object,
	Shape extends ModelShape, Identifier extends keyof Shape = any, TSchema extends TablesRelationalConfig = ExtractTablesWithRelations<TFullSchema>
>
{
	/**
	 * Drizzle model manager.
	 */
	drizzle(): DrizzleModel<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>;
}

/**
 * Drizzle model manager.
 */
export class DrizzleModel<
	ModelType extends Model<Shape, IdentifierType<Shape, Identifier>>,
	TableType extends Table<TC>, TC extends TableConfig,
	TQueryResult extends PgQueryResultHKT, TFullSchema extends Record<string, unknown>,
	Scopes extends object,
	Shape extends ModelShape, Identifier extends keyof Shape = any, TSchema extends TablesRelationalConfig = ExtractTablesWithRelations<TFullSchema>
>
{
	constructor(protected modelClass: ModelClass<ModelType & DrizzleExtension<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>, Shape, Identifier>,
	            protected model: ModelType,
	            protected database: PgDatabase<TQueryResult, TFullSchema, TSchema>,
	            protected table: TableType,
	            protected scopes: Scopes & ThisType<ModelQuery<ModelType, TableType, TC, TQueryResult, TFullSchema, Shape, Identifier, TSchema>>)
	{}

	/**
	 * Get Drizzle database.
	 */
	getDatabase(): PgDatabase<TQueryResult, TFullSchema, TSchema>
	{
		return this.database;
	}

	/**
	 * Get model Drizzle table.
	 */
	getTable(): TableType
	{
		return this.table;
	}

	/**
	 * Get scopes.
	 */
	getScopes(): Scopes & ThisType<ModelQuery<ModelType, TableType, TC, TQueryResult, TFullSchema, Shape, Identifier, TSchema>>
	{
		return this.scopes;
	}

	/**
	 * Save the model in database.
	 */
	async save(): Promise<boolean>
	{
		// New query to save the model.
		return query(this.modelClass).save(this.model as any);
	}

	/**
	 * Refresh model data from database.
	 */
	async refresh(): Promise<void>
	{
		// New query to refresh the model.
		await query(this.modelClass).refresh(this.model as any);
	}
}

/**
 * Drizzle extension initializer.
 * @param database Drizzle database.
 * @param table Drizzle database table.
 * @param scopes
 */
export function drizzle<
	ModelType extends Model<Shape, IdentifierType<Shape, Identifier>>,
	TableType extends Table<TC>, TC extends TableConfig,
	TQueryResult extends PgQueryResultHKT, TFullSchema extends Record<string, unknown>,
	Scopes extends object,
	Shape extends ModelShape, Identifier extends keyof Shape = any, TSchema extends TablesRelationalConfig = ExtractTablesWithRelations<TFullSchema>,
>(
	database: PgDatabase<TQueryResult, TFullSchema, TSchema>,
	table: TableType,
	scopes: Scopes & ThisType<ModelQuery<ModelType, TableType, TC, TQueryResult, TFullSchema, Shape, Identifier, TSchema>>
): DrizzleExtension<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema> & ThisType<ModelType>
{
	// Return initialized drizzle extension.
	return {
		drizzle(): DrizzleModel<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>
		{
			// Initialize drizzle model manager instance.
			return new DrizzleModel<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>(
				this.constructor as ModelClass<ModelType & DrizzleExtension<ModelType, TableType, TC, TQueryResult, TFullSchema, Scopes, Shape, Identifier, TSchema>, Shape, Identifier>,
				this, database, table, scopes
			);
		}
	};
}
